﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;

namespace HomeFinance.Model
{
    static class CalcoloFactory
    {
        private static readonly Dictionary<string, ICalcolo> _dictionary = new Dictionary<string, ICalcolo>();

        static CalcoloFactory()
        {
            //Add(new CalcolaSomma());
            //Add(new CalcolaMedia());
            //Add(new CalcolaCount());

            foreach (Type type in Assembly.GetExecutingAssembly().GetTypes())
            {
                if (type.GetInterface(typeof(ICalcolo).Name) != null && !type.IsAbstract)
                {
                    Add((ICalcolo) Activator.CreateInstance(type));
                }
            }
        }

        public static ICalcolo GetCalcolo(string nome)
        {
            if (!_dictionary.ContainsKey(nome))
                throw new ArgumentException("!_dictionary.ContainsKey(nome)");
            return _dictionary[nome];
        }

        public static IEnumerable<string> GetNomi()
        {
            return _dictionary.Keys;
        }

        public static IEnumerable<ICalcolo> GetCalcoli()
        {
            return _dictionary.Values;
        }

        private static void Add(ICalcolo calcolo)
        {
            if (calcolo == null)
                throw new ArgumentNullException("calcolo");
            if (_dictionary.ContainsKey(calcolo.Nome))
                throw new ArgumentException("_dictionary.ContainsKey(calcolo.Nome)");
            _dictionary.Add(calcolo.Nome, calcolo);
        }
    }

    abstract class CalcoloBase : ICalcolo
    {
        private readonly string _nome;

        protected CalcoloBase(string nome)
        {
            if (String.IsNullOrEmpty(nome))
                throw new ArgumentException("String.IsNullOrEmpty(nome)");
            _nome = nome;
        }

        public string Nome
        {
            get { return _nome; }
        }

        public abstract double Calcola(IEnumerable<Movimento> movimenti);
    }

    class CalcolaSomma : CalcoloBase
    {
        public CalcolaSomma()
            : base("Somma")
        {
        }

        public override double Calcola(IEnumerable<Movimento> movimenti)
        {
            return movimenti.Sum(movimento => movimento.Importo);
        }
    }

    class CalcolaMedia : CalcoloBase
    {
        public CalcolaMedia()
            : base("Media")
        {
        }

        public override double Calcola(IEnumerable<Movimento> movimenti)
        {
            if (movimenti.Count() > 0)
                return movimenti.Average(movimento => movimento.Importo);
            else
                return 0.0;
        }
    }

    class CalcolaCount : CalcoloBase
    {
        public CalcolaCount()
            : base("Count")
        {
        }

        public override double Calcola(IEnumerable<Movimento> movimenti)
        {
            return movimenti.Count();
        }
    }
}