﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace HomeFinance.Model
{
    abstract class SelezionatoreBase : ISelezionatore
    {
        private readonly ISelezionatore _selezionatore;

        public readonly static ISelezionatore SelezionatoreDiTutto = new _SelezionatoreDiTutto();
        public readonly static ISelezionatore SelezionatoreDiNiente = new _SelezionatoreDiNiente();

        protected SelezionatoreBase(ISelezionatore selezionatore)
        {
            _selezionatore = selezionatore ?? SelezionatoreDiTutto;
        }

        public IEnumerable<MovimentoReale> GetMovimenti()
        {
            return _selezionatore.GetMovimenti().Where(Predicate);
        }

        public override string ToString()
        {
            if (_selezionatore is SelezionatoreBase)
                return Description + " + " + _selezionatore.ToString();
            else
                return Description;
        }

        public abstract string Description { get; }

        protected abstract Func<MovimentoReale, bool> Predicate { get; }

        private class _SelezionatoreDiTutto : ISelezionatore
        {
            public IEnumerable<MovimentoReale> GetMovimenti()
            {
                return Documento.GetInstance().GetMovimentiReali();
            }

            public override string ToString()
            {
                return "tutto";
            }
        }

        private class _SelezionatoreDiNiente : ISelezionatore
        {
            public IEnumerable<MovimentoReale> GetMovimenti()
            {
                return new MovimentoReale[0];
            }

            public override string ToString()
            {
                return "niente";
            }
        }
    }

    class SelezionatorePerCategorie : SelezionatoreBase
    {
        private readonly IEnumerable<Categoria> _categorie;

        public SelezionatorePerCategorie(ISelezionatore selezionatore, params Categoria[] categorie)
            : this(selezionatore, (IEnumerable<Categoria>) categorie)
        {
        }

        public SelezionatorePerCategorie(ISelezionatore selezionatore, IEnumerable<Categoria> categorie)
            : base(selezionatore)
        {
            if (categorie == null)
                throw new ArgumentNullException("categorie");
            if (categorie.Contains(null))
                throw new ArgumentException("categorie.Contains(null)");
            _categorie = categorie;
        }

        public override string Description
        {
            get
            {
                StringBuilder sb = new StringBuilder();
                foreach (Categoria categoria in _categorie)
                {
                    if (sb.Length > 0)
                        sb.Append(", ");
                    sb.Append(categoria.Nome);
                }
                return "categorie { " + sb.ToString() + " }";
            }
        }

        protected override Func<MovimentoReale, bool> Predicate
        {
            get { return movimento => _categorie.Contains(movimento.Categoria); }
        }
    }

    class SelezionatoreDallaData : SelezionatoreBase
    {
        private readonly DateTime _dallaData;

        public SelezionatoreDallaData(ISelezionatore selezionatore, DateTime dallaData)
            : base(selezionatore)
        {
            _dallaData = dallaData.Date;
        }

        public override string Description
        {
            get { return "dal giorno " + _dallaData.ToString("d"); }
        }

        protected override Func<MovimentoReale, bool> Predicate
        {
            get { return movimento => _dallaData <= movimento.Date; }
        }
    }

    class SelezionatoreAllaData : SelezionatoreBase
    {
        private readonly DateTime _allaData;

        public SelezionatoreAllaData(ISelezionatore selezionatore, DateTime allaData)
            : base(selezionatore)
        {
            _allaData = allaData.Date;
        }

        public override string Description
        {
            get { return "al giorno " + _allaData.ToString("d"); }
        }

        protected override Func<MovimentoReale, bool> Predicate
        {
            get { return movimento => movimento.Date <= _allaData; }
        }
    }

    class SelezionatorePerTipoMovimento : SelezionatoreBase
    {
        private readonly TipoMovimento _tipoMovimento;

        public SelezionatorePerTipoMovimento(ISelezionatore selezionatore, TipoMovimento tipoMovimento)
            : base(selezionatore)
        {
            if (tipoMovimento != TipoMovimento.Ordinario && tipoMovimento != TipoMovimento.Straordinario)
                throw new ArgumentException("TipoMovimento errato");
            _tipoMovimento = tipoMovimento;
        }

        public override string Description
        {
            get { return "tipo movimento " + _tipoMovimento.ToString(); }
        }

        protected override Func<MovimentoReale, bool> Predicate
        {
            get { return movimento => movimento.Tipo == _tipoMovimento; }
        }
    }
}
